import pickle

import numpy as np
from sklearn.model_selection import train_test_split


class Emulator:
    def __init__(
        self,
        test_size=0.2,
        preconditioner=None,
    ):
        """
        Base class for emulators

        Parameters
        ----------
        test_size : float
            Fraction of samples to use for testing
        preconditioner : object
            Implement forward and backward methods for transforming function results before
            training and after prediction
        """
        self.test_size = test_size
        self.preconditioner = preconditioner  # need to have forward and backward methods for transf and inverse transf
        self.model = None

        # input, output dimensions, and parameter names
        self.idim = None
        self.odim = None
        self.param_keys = None  # TODO: retrieve from function signature instead

        self.X_test = None
        self.Y_test = None

    def _prepare_data(self, func, samples, precomputed=None, collection=None):
        """
        Prepare data for training and testing

        Parameters
        ----------
        func : callable
            Function to be emulated
        samples : dict
            Dictionary of parameter names and values
        precomputed: array-like
            Precomputed outputs of the function. If provided, the function will not be
            evaluated.
        collection: str
            Path to a pickle file containing precomputed collection of data. If
            provided, the function will not be evaluated. It takes precedence over
            precomputed which contains only the function outputs. The collection
            is expected to be the saved product of the `collect` function.

        Returns
        -------
        X_train : np.ndarray
            Training input
        X_test : np.ndarray
            Test input
        Y_train : np.ndarray
            Training output
        Y_test : np.ndarray
            Test output
        """
        # load everything from collection if it's given
        if collection is not None:
            with open(collection, "rb") as f:
                collection_ = pickle.load(f)
            # use the loaded values instead of computing
            X = collection_["X"]
            Y = collection_["Y"]
            samples = collection_["samples"]
        else:
            X = np.vstack([np.array(v) for _, v in samples.items()]).T

            # run all combinations and gather results
            if precomputed is None:
                res_list = []
                for i in range(
                    X.shape[0]
                ):  # samples has shape (n_samples, n_parameters)
                    res = func(**{k: v for k, v in zip(samples.keys(), X[i, :])})
                    res_list.append(res)
                Y = np.array(res_list)
            else:
                Y = np.array(precomputed)

        # optionally apply a transformation
        if self.preconditioner is not None:
            Y = self.preconditioner.forward(Y)

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        if len(Y.shape) == 1:
            Y = Y.reshape(-1, 1)

        # split into train and test
        X_train, X_test, Y_train, Y_test = train_test_split(
            X, Y, test_size=self.test_size
        )

        # save input, output dimensions, and parameter names
        self.idim = X_train.shape[1]
        self.odim = Y_train.shape[1]
        self.param_keys = list(
            samples.keys()
        )  # TODO: retrieve from function signature instead

        # save a reference to test data for future use
        self.X_test = X_test
        self.Y_test = Y_test

        return X_train, X_test, Y_train, Y_test

    def _train(self, X_train, Y_train):
        """
        Train the emulator

        Parameters
        ----------
        X_train : np.ndarray
            Training input
        Y_train : np.ndarray
            Training output
        """
        raise NotImplementedError

    def _test(self, X_test, Y_test):
        """
        Test the emulator

        Parameters
        ----------
        X_test : np.ndarray
            Test input
        Y_test : np.ndarray
            Test output
        """
        raise NotImplementedError

    def _predict(self, x):
        """
        Predict the output for a given input

        Parameters
        ----------
        x : np.ndarray
            Input parameters for emulated function

        Returns
        -------
        y : np.ndarray
            Output of emulated function
        """
        raise NotImplementedError

    def _build_func(self):
        """
        Build a function that can be used to emulate the original function

        Parameters
        ----------
        param_keys : list of str
            List of parameter names

        Returns
        -------
        emulated_func : callable
            Function that can be used to emulate the original function. The function has a
            reference to the emulator instance in the attribute `emulator`.
        """

        def emulated_func(*args, **kwargs):
            # check if the function has been called with the correct number of arguments
            if len(args) > 0:
                if len(args) != len(self.param_keys):
                    raise ValueError("Incorrect number of arguments")
                values = np.array(args)
            else:
                values = [kwargs.get(k, None) for k in self.param_keys]
                if None in values:
                    raise ValueError("Missing argument")

            res = self._predict(values)
            if self.preconditioner is not None:
                res = self.preconditioner.backward(res)
            return res

        # attach a reference to the emulator instance in the function
        emulated_func.emulator = self
        return emulated_func

    def emulate(
        self, func, samples, pretrained=None, precomputed=None, collection=None
    ):
        """
        Emulate a function

        Parameters
        ----------
        func : callable
            Function to be emulated
        samples : dict
            Dictionary of parameter names and values, where the values are lists of
            parameter values to evaluate the function at to build the emulator
        precomputed: array-like
            Precomputed outputs of the function. If provided, the function will not be
            evaluated at the parameter values in `samples`

        Returns
        -------
        emulated_func : callable
            Function that can be used to emulate the original function. The function has a
            reference to the emulator instance in the attribute `emulator`.
        """
        if pretrained is not None:
            self.load(pretrained)
            return self._build_func()
        X_train, X_test, Y_train, Y_test = self._prepare_data(
            func, samples, precomputed=precomputed, collection=collection
        )
        self._train(X_train, Y_train)
        self._test(X_test, Y_test)
        emulated_func = self._build_func()
        return emulated_func

    def _save(self, store):
        """
        Save the emulator to a datastore

        Parameters
        ----------
        store : dict
            DataStore to save the emulator to
        """
        raise NotImplementedError

    def _load(self, store):
        """
        Load the emulator from a datastore

        Parameters
        ----------
        store : dict
            DataStore to load the emulator from
        """
        raise NotImplementedError

    def save(self, filename):
        """
        Save the emulator to a file

        Parameters
        ----------
        filename : str
            Filename to save the emulator to
        """
        # always save the input and output dimensions
        store = {
            "in": self.idim,
            "out": self.odim,
            "param_keys": self.param_keys,
        }
        self._save(store)
        if self.preconditioner is not None:
            if not hasattr(self.preconditioner, "_save"):
                raise ValueError("Cannot save a preconditioner without a _save method")
            store["preconditioner"] = {}
            self.preconditioner._save(store["preconditioner"])

        with open(filename, "wb") as f:
            pickle.dump(store, f)

    def load(self, filename):
        """
        Load the emulator from a file

        Parameters
        ----------
        filename : str
            Filename to load the emulator from
        """
        with open(filename, "rb") as f:
            store = pickle.load(f)

        self.idim = store["in"]
        self.odim = store["out"]
        self.param_keys = store["param_keys"]

        if "preconditioner" in store:
            if self.preconditioner is None:
                raise ValueError(
                    "Cannot load a preconditioner into an emulator without one"
                )
            self.preconditioner._load(store["preconditioner"])
        self._load(store)


# the main emulator interface
class emulate:
    def __init__(
        self,
        samples=None,
        emulator_class=None,
        backend="jax",
        save=None,
        load=None,
        precomputed=None,
        collection=None,
        **kwargs,
    ):
        """
        A class used to emulate a function using a machine learning model.

        Parameters
        ----------
        samples : np.ndarray or pd.DataFrame, optional
            The samples to use for training the machine learning model. If None, the emulator will not be trained and cannot be used for emulation.
        emulator_class : type, optional
            The class of the emulator to use. If None, the default emulator for the specified backend will be used.
        backend : str, optional
            The backend library to use for the emulator. Currently supported options are "torch", "sklearn", "gpy", "jax". Default is "jax".
        save : str, optional
            The path to the file to save the emulator to. If None, the emulator will not be saved. Default is None.
        load : str, optional
            The path to the file to load the emulator from. If None, the emulator will not be loaded. Default is None.
        precomputed : np.ndarray, optional
            precomputed outputs of the function. If provided, the function will not be evaluated at the parameter values
            in `samples`.
        collection : str, optional
            A dictionary of precomputed collection of the function. If provided, the function will not be evaluated
            at the parameter values in `samples`, and `X`, `Y`, `samples` stored in `collection` dictionary will be used
            instead. This works with collection saved by the `collect` decorator.
        **kwargs
            Additional keyword arguments to pass to the emulator constructor.

        Attributes
        ----------
        samples : np.ndarray or pd.DataFrame
            The samples used for training the machine learning model.
        emulator_class : type
            The class of the emulator used for emulation.
        em : emulator_class
            An instance of the emulator used for emulation.
        save : str
            The path to the file to save the emulator to.
        load : str
            The path to the file to load the emulator from.

        Methods
        -------
        __call__(func)
            Call the emulator to emulate the specified function with the specified input samples.
        """

        self.samples = samples
        self.precomputed = precomputed
        self.save = save
        self.load = load
        self.collection = collection

        # select the backend
        if backend == "torch":
            try:
                from .backends.torch import TorchEmulator
            except ImportError:
                raise ImportError(
                    "PyTorch is not installed. Please install it to use the PyTorch backend."
                )
            self.emulator_class = TorchEmulator
        elif backend == "sklearn":
            from .backends.sklearn import SklearnEmulator  # installed by default

            self.emulator_class = SklearnEmulator
        elif backend == "gpy":
            try:
                from .backends.GPy import GPyEmulator
            except ImportError:
                raise ImportError(
                    "GPy is not installed. Please install it to use the GPy backend."
                )
            self.emulator_class = GPyEmulator
        elif backend == "jax":
            try:
                from .backends.jax import JaxEmulator
            except ImportError:
                raise ImportError(
                    "Jax is not installed. Please install jax, optax, and flax to use the Jax backend."
                )
            self.emulator_class = JaxEmulator
        else:
            raise ValueError(f"Unknown backend {backend}")

        # if custom emulator class is provided, use it instead
        if emulator_class is not None:
            self.emulator_class = emulator_class

        self.em = self.emulator_class(**kwargs)

    def __call__(self, func):
        """
        Call the emulator to emulate the specified function with the specified input samples.

        Parameters
        ----------
        func : callable
            The function to emulate.

        Returns
        -------
        callable
            The emulator function.
        """
        emulated = self.em.emulate(
            func,
            self.samples,
            pretrained=self.load,
            precomputed=self.precomputed,
            collection=self.collection,
        )
        if self.save is not None:
            self.em.save(self.save)
        return emulated



import os
import pickle

import numpy as np

from galilei import emulate
from galilei.experimental import collect
from galilei.sampling import build_samples


def test_collect():
    @collect(samples=build_samples({"a": [0, 2], "b": [0, 2]}, 100), save="_test.pkl")
    def test_em(a, b):
        x = np.linspace(0, 10, 100)
        return np.sin(a * x) + np.sin(b * x)

    @emulate(collection="_test.pkl", backend="sklearn")
    def test_em(a, b):
        pass

    # if collection loaded correctly, this should work
    output = test_em(1, 1)
    assert output.shape == (100,)

    # clean up
    if os.path.exists("_test.pkl"):
        os.remove("_test.pkl")



from functools import partial

import torch
from torch import nn, optim
from torch.utils.data import DataLoader, TensorDataset
from tqdm import tqdm

from ..emulator import Emulator


class Net(nn.Module):
    def __init__(self, idim, odim, hidden=[128, 128], activation=torch.sigmoid):
        """
        Simple fully connected neural network with ReLU activation

        Parameters
        ----------
        idim : int
            Input dimension
        odim : int
            Output dimension
        hidden : list of int
            List of hidden layer sizes
        """
        super(Net, self).__init__()
        self.activation = activation
        self.layers = nn.ModuleList()
        self.layers.append(nn.Linear(idim, hidden[0]))
        for i in range(1, len(hidden)):
            self.layers.append(nn.Linear(hidden[i - 1], hidden[i]))
        self.layers.append(nn.Linear(hidden[-1], odim))

    def forward(self, out):
        for layer in self.layers[:-1]:
            out = self.activation(layer(out))
        out = self.layers[-1](out)
        return out


class TorchEmulator(Emulator):
    def __init__(
        self,
        NNClass=Net,
        epochs=500,
        optimizer=None,
        criterion=None,
        lr=0.01,
        weight_decay=0,
        momentum=0.9,
        NN_kwargs={},
        scheduler=None,
        batch_size=32,
        device=None,
        **kwargs,
    ):
        """
        A class used to emulate data using a neural network based on the PyTorch library.

        Parameters
        ----------
        NNClass : nn.Module, optional
            The neural network class to use. Default is Net.
        epochs : int, optional
            The number of epochs to train the neural network. Default is 500.
        optimizer : torch.optim.Optimizer, optional
            The optimizer to use for training the neural network. If None, uses Stochastic Gradient Descent (SGD) optimizer with default parameters.
        criterion : torch.nn.modules.loss._Loss, optional
            The loss function to use for training the neural network. If None, uses Mean Squared Error (MSE) loss function.
        lr : float, optional
            The learning rate for the optimizer. Default is 0.01.
        weight_decay : float, optional
            The weight decay (L2 penalty) for the optimizer. Default is 0.
        momentum : float, optional
            The momentum factor for the optimizer. Default is 0.9.
        NN_kwargs : dict, optional
            The keyword arguments to pass to the neural network class constructor. Default is an empty dictionary.
        **kwargs
            Additional keyword arguments to pass to the base Emulator class constructor.

        Attributes
        ----------
        NNClass : callable
            A callable that returns an instance of the neural network class with the given keyword arguments.
        epochs : int
            The number of epochs to train the neural network.
        optimizer : torch.optim.Optimizer
            The optimizer to use for training the neural network.
        criterion : torch.nn.modules.loss._Loss
            The loss function to use for training the neural network.
        """
        super(TorchEmulator, self).__init__(**kwargs)
        self.NNClass = partial(NNClass, **NN_kwargs)
        self.epochs = epochs
        self.optimizer = (
            optimizer
            if optimizer is not None
            else partial(optim.SGD, lr=lr, momentum=momentum, weight_decay=weight_decay)
        )
        self.criterion = criterion if criterion is not None else nn.MSELoss()
        self.scheduler = scheduler
        self.batch_size = batch_size
        if device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device

    def _train(self, X_train, Y_train):
        """
        Train the emulator

        Parameters
        ----------
        X_train : np.ndarray
            Training input
        Y_train : np.ndarray
            Training output
        """
        device = self.device
        # convert the data to PyTorch tensors and send to GPU if available
        X_train = torch.as_tensor(X_train, dtype=torch.float32, device=device)
        Y_train = torch.as_tensor(Y_train, dtype=torch.float32, device=device)

        # create TensorDatasets and DataLoaders
        train_data = TensorDataset(X_train, Y_train)
        train_loader = DataLoader(train_data, batch_size=32, shuffle=True)

        # build model, optimizer and loss function
        model = self.NNClass(X_train.shape[-1], Y_train.shape[-1]).to(device)
        optimizer = self.optimizer(model.parameters())
        scheduler = self.scheduler(optimizer) if self.scheduler is not None else None

        # train the model
        print("Training emulator...")
        pbar = tqdm(range(self.epochs))
        for epoch in pbar:
            running_loss = 0.0
            for i, (inputs, Y) in enumerate(train_loader):
                # zero the parameter gradients
                optimizer.zero_grad()

                # forward + backward + optimize
                outputs = model(inputs)
                loss = self.criterion(outputs, Y)
                loss.backward()
                optimizer.step()

                # print statistics
                running_loss += loss.item()
            pbar.set_postfix({"loss": f"{running_loss/(i+1):.5f}"})
            if scheduler is not None:
                scheduler.step()

        self.model = model

    def _test(self, X_test, Y_test):
        """
        Test the emulator on a test set

        Parameters
        ----------
        X_test : array-like
            The test set input data
        Y_test : array-like
            The test set output data
        """
        device = self.device
        X_test = torch.as_tensor(X_test, dtype=torch.float32, device=device)
        Y_test = torch.as_tensor(Y_test, dtype=torch.float32, device=device)
        # Test the model
        if len(X_test) > 0:
            with torch.no_grad():
                test_loss = 0.0
                for i, data in enumerate(zip(X_test, Y_test)):
                    inputs, Y = data
                    outputs = self.model(inputs)
                    loss = self.criterion(outputs, Y)
                    test_loss += loss.item()
                avg_test_loss = test_loss / (i + 1)
            print("Ave Test loss: {:.3f}".format(avg_test_loss))

    def _predict(self, x):
        """
        Predict the output of the emulator for a given input

        Parameters
        ----------
        x : array-like
            The input data

        Returns
        -------
        np.ndarray
            The predicted output
        """
        x = torch.as_tensor(x, dtype=torch.float32, device=self.device)
        res = self.model(x).detach().cpu().numpy()
        return res

    def _save(self, store):
        """
        Save the emulator model to a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk

        """
        store["model"] = self.model.state_dict()

    def _load(self, store):
        """
        Load the emulator model from a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk
        """
        self.model = self.NNClass(store["in"], store["out"])
        self.model.load_state_dict(store["model"])
        self.model.to(self.device)



"""Console script for galilei."""

import fire


def help():
    print("galilei")
    print("=" * len("galilei"))
    print("the galilei project")


def main():
    fire.Fire({"help": help})


if __name__ == "__main__":
    main()  # pragma: no cover



from collections import OrderedDict

import numpy as np


def build_samples(sample_dict, nsamps, method="lhs"):
    """Build samples from a sample dictionary.

    Parameters
    ----------
    sample_dict : dict
        Dictionary of samples like {key: [min, max]}
    nsamps : int
        Number of samples to build.
    method : str, optional
        Method to use for building samples. The default is 'lhs'.

    Returns
    -------
    samples : dict
        Dictionary of samples.

    """
    # validation
    for key, value in sample_dict.items():
        if len(value) != 2:
            raise ValueError(
                "Sample dictionary values must be of length 2, for min and max!"
            )
        assert value[0] < value[1], "Min must be less than max!"
    assert nsamps > 0, "Number of samples must be greater than 0!"

    # build samples
    if method == "lhs":
        design_func = build_latin_hypercube
    else:
        return NotImplementedError("Only Latin hypercube sampling is implemented!")
    nparams = len(sample_dict)
    design_matrix = design_func(
        nsamps, nparams
    )  # n x m matrix with each element from 0 to 1

    # rescale design matrix to match the desired range given as min and max
    samples = OrderedDict()
    for i, key in enumerate(sample_dict.keys()):
        samples[key] = sample_dict[key][0] + design_matrix[:, i] * (
            sample_dict[key][1] - sample_dict[key][0]
        )
    return samples


# utility functions
def build_latin_hypercube(n, m):
    """Build a Latin hypercube of size n x m.

    Parameters
    ----------
    n : int
        Number of rows.
    m : int
        Number of columns.

    Returns
    -------
    latin_hypercube : array_like
        Latin hypercube of size n x m.

    """
    latin_hypercube = np.zeros((n, m))
    for i in range(m):
        latin_hypercube[:, i] = np.random.permutation(n)
    design_matrix = np.zeros((n, m))
    for i in range(m):
        design_matrix[:, i] = (latin_hypercube[:, i] + np.random.uniform(size=n)) / n
    return design_matrix



import GPy
import numpy as np

from ..emulator import Emulator


class GPyEmulator(Emulator):
    def __init__(
        self,
        kernel_kwargs={},
        model_kwargs={},
        **kwargs,
    ):
        """
        A class used to emulate data using a machine learning model based on the GPy library.

        Parameters
        ----------
        kernel_kwargs : dict, optional
            The keyword arguments to pass to the RBF kernel constructor if a Gaussian Process model is used. Default is an empty dictionary.
        model_kwargs : dict, optional
            The keyword arguments to pass to the machine learning model constructor if one is provided. Default is an empty dictionary.
        **kwargs
            Additional keyword arguments to pass to the base Emulator class constructor.

        Attributes
        ----------
        model : GPy.models.GPRegression
            The machine learning model used for emulation.
        """
        super(GPyEmulator, self).__init__(**kwargs)
        self.kernel_kwargs = kernel_kwargs
        self.model_kwargs = model_kwargs
        self.model = None

    def _train(self, X_train, Y_train):
        """
        Train the emulator

        Parameters
        ----------
        X_train : np.ndarray
            Training input
        Y_train : np.ndarray
            Training output
        """
        kernel = GPy.kern.RBF(input_dim=X_train.shape[-1], **self.kernel_kwargs)
        model = GPy.models.GPRegression(X_train, Y_train, kernel, **self.model_kwargs)
        model.optimize()
        self.model = model

    def _test(self, X_test, Y_test):
        """
        Test the emulator on a test set

        Parameters
        ----------
        X_test : array-like
            The test set input data
        Y_test : array-like
            The test set output data
        """
        if len(X_test) > 0:
            test_loss = np.mean((Y_test - self.model.predict(X_test)[0]) ** 2)
            print("Ave Test loss: {:.5f}".format(test_loss))

    def _predict(self, x):
        """
        Predict the output of the emulator for a given input

        Parameters
        ----------
        x : array-like
            The input data

        Returns
        -------
        np.ndarray
            The predicted output
        """
        res = self.model.predict(np.array([x]))
        return res[0][0]

    def _save(self, store):
        """
        Save the emulator model to a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk
        """
        store["model"] = self.model.to_dict()

    def _load(self, store):
        """
        Load the emulator model from a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk
        """
        import GPy

        self.model = GPy.models.GPRegression.from_dict(store["model"])



import pickle

import numpy as np


class collect:
    def __init__(self, samples, save="data.pkl", mpi=False):
        """
        collect data from a function

        Parameters
        ----------
        samples: dict
            dictionary of samples
        save: str
            path to save the data
        mpi: bool
            use mpi to collect data, default False, only works when
            mpi4py is installed
        """
        self.samples = samples
        self.mpi = mpi
        self.save = save

        if self.mpi:
            try:
                from mpi4py import MPI

                self.comm = MPI.COMM_WORLD
                self.rank = self.comm.Get_rank()
                self.size = self.comm.Get_size()
            except ImportError:
                raise ImportError(
                    "mpi4py is not installed, " "please install it to use mpi"
                )
        else:
            self.comm = None
            self.rank = 0
            self.size = 1

    def __call__(self, func):
        """
        collect data from a function

        Parameters
        ----------
        func: function
            function to collect data from
        """
        X = np.vstack([np.array(v) for _, v in self.samples.items()]).T
        res_list = []
        for i in range(self.rank, X.shape[0], self.size):
            res = func(**{k: v for k, v in zip(self.samples.keys(), X[i, :])})
            res_list.append(res)
        Y = np.array(res_list)

        if self.mpi:
            Y = self.comm.gather(Y, root=0)
            if self.rank == 0:
                Y = np.vstack(Y)

        if self.rank == 0:
            if len(X.shape) == 1:
                X = X.reshape(-1, 1)
            if len(Y.shape) == 1:
                Y = Y.reshape(-1, 1)

            # save output
            to_save = {
                "samples": self.samples,
                "X": X,
                "Y": Y,
            }
            with open(self.save, "wb") as f:
                pickle.dump(to_save, f)



#!/usr/bin/env python
"""Tests for `galilei.sample` package."""
# pylint: disable=redefined-outer-name
import numpy as np
import pytest

from galilei import sampling


def test_build_latin_hypercube():
    # Test case 1: Test if the returned array is of the correct size
    n = 5
    m = 4
    result = sampling.build_latin_hypercube(n, m)
    assert result.shape == (n, m)

    # Test case 2: Test if all elements of the returned array are between 0 and 1
    assert np.all(result >= 0) and np.all(result <= 1)

    # Test case 3: Test if all columns of the returned array are unique
    for i in range(m):
        assert len(set(result[:, i])) == n


def test_build_samples():
    # Sample data
    sample_dict = {"x1": [0, 1], "x2": [-1, 2], "x3": [5, 10]}
    nsamps = 10

    # Test case 1: Test if the returned dictionary has the correct number of keys
    result = sampling.build_samples(sample_dict, nsamps)
    assert len(result.keys()) == len(sample_dict.keys())

    # Test case 2: Test if the returned dictionary has the correct number of values in each key
    for key in result.keys():
        assert len(result[key]) == nsamps

    # Test case 3: Test if the returned dictionary values are within the correct range for each key
    for key in sample_dict.keys():
        assert np.all(result[key] >= sample_dict[key][0]) and np.all(
            result[key] <= sample_dict[key][1]
        )

    # Test case 4: Test if ValueError is raised when sample_dict values are not of length 2
    with pytest.raises(ValueError):
        sample_dict = {"x1": [0, 1, 2], "x2": [-1, 2], "x3": [5, 10]}
        sampling.build_samples(sample_dict, nsamps)

    # Test case 5: Test if AssertionError is raised when min is greater than or equal to max
    with pytest.raises(AssertionError):
        sample_dict = {"x1": [1, 0], "x2": [-1, 2], "x3": [5, 10]}
        sampling.build_samples(sample_dict, nsamps)

    # Test case 6: Test if AssertionError is raised when nsamps is less than or equal to 0
    with pytest.raises(AssertionError):
        sampling.build_samples(sample_dict, 0)



from sklearn.gaussian_process import GaussianProcessRegressor, kernels
from sklearn.metrics import mean_squared_error

from ..emulator import Emulator


class SklearnEmulator(Emulator):
    def __init__(
        self,
        model=None,
        kernel_kwargs={},
        model_kwargs={"n_restarts_optimizer": 10},
        **kwargs,
    ):
        """
        A class used to emulate data using a machine learning model based on the scikit-learn library.

        Parameters
        ----------
        model : sklearn.base.TransformerMixin, optional
            The machine learning model to use. If None, uses a Gaussian Process model with an RBF kernel from scikit-learn with default parameters.
        kernel_kwargs : dict, optional
            The keyword arguments to pass to the RBF kernel constructor if a Gaussian Process model is used. Default is an empty dictionary.
        model_kwargs : dict, optional
            The keyword arguments to pass to the machine learning model constructor if one is provided. Default sets n_restarts_optimizer to 5.
        **kwargs
            Additional keyword arguments to pass to the base Emulator class constructor.

        Attributes
        ----------
        model : sklearn.base.TransformerMixin
            The machine learning model used for emulation.
        """

        super(SklearnEmulator, self).__init__(**kwargs)
        if model is None:
            # use a gaussian process model with RKB kernel from sklearn if none is provided
            model = GaussianProcessRegressor(
                kernel=kernels.RBF(**kernel_kwargs), **model_kwargs
            )
        self.model = model

    def _train(self, X_train, Y_train):
        """
        Train the emulator

        Parameters
        ----------
        X_train : np.ndarray
            Training input
        Y_train : np.ndarray
            Training output
        """
        self.model.fit(X_train, Y_train)

    def _test(self, X_test, Y_test):
        """
        Test the emulator on a test set

        Parameters
        ----------
        X_test : array-like
            The test set input data
        Y_test : array-like
            The test set output data
        """
        if len(X_test) > 0:
            test_loss = mean_squared_error(Y_test, self.model.predict(X_test))
            print("Ave Test loss: {:.3f}".format(test_loss))

    def _predict(self, x):
        """
        Predict the output of the emulator for a given input

        Parameters
        ----------
        x : array-like
            The input data

        Returns
        -------
        np.ndarray
            The predicted output
        """
        res = self.model.predict([x])[0]
        return res

    def _save(self, store):
        """
        Save the emulator model to a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk

        """
        store["model"] = self.model

    def _load(self, store):
        """
        Load the emulator model from a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk
        """
        self.model = store["model"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['galilei', 'galilei.backends', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['fire==0.4.0',
 'flax>=0.6.4,<0.7.0',
 'jaxlib>=0.4.4,<0.5.0',
 'optax>=0.1.4,<0.2.0',
 'scikit-learn>=1.2.1,<2.0.0',
 'tqdm>=4.65.0,<5.0.0']

extras_require = \
{'dev': ['tox>=3.24.5,<4.0.0',
         'virtualenv>=20.13.1,<21.0.0',
         'pip>=22.0.3,<23.0.0',
         'twine>=3.8.0,<4.0.0',
         'pre-commit>=2.17.0,<3.0.0',
         'toml>=0.10.2,<0.11.0'],
 'doc': ['mkdocs>=1.2.3,<2.0.0',
         'mkdocs-include-markdown-plugin>=3.2.3,<4.0.0',
         'mkdocs-material>=8.1.11,<9.0.0',
         'mkdocstrings>=0.18.0,<0.19.0',
         'mkdocs-autorefs>=0.3.1,<0.4.0',
         'mike>=1.1.2,<2.0.0'],
 'test': ['black>=22.3.0,<23.0.0',
          'isort==5.10.1',
          'flake8==4.0.1',
          'flake8-docstrings>=1.6.0,<2.0.0',
          'pytest>=7.0.1,<8.0.0',
          'pytest-cov>=3.0.0,<4.0.0']}

entry_points = \
{'console_scripts': ['galilei = galilei.cli:main']}

setup_kwargs = {
    'name': 'galilei',
    'version': '0.3.1',
    'description': 'the galilei project.',
    'long_description': '# galilei\n<a href="https://pypi.python.org/pypi/galilei">\n    <img src="https://img.shields.io/pypi/v/galilei.svg"\n        alt = "Release Status">\n</a>\n<a href="https://github.com/guanyilun/galilei/actions">\n    <img src="https://github.com/guanyilun/galilei/actions/workflows/release.yml/badge.svg?branch=master" alt="CI Status">\n</a>\n<a href="https://github.com/guanyilun/galilei/actions">\n    <img src="https://github.com/guanyilun/galilei/actions/workflows/dev.yml/badge.svg?branch=master" alt="CI Status">\n</a>\n<a href="https://guanyilun.github.io/galilei/">\n    <img src="https://img.shields.io/website/https/guanyilun.github.io/galilei/index.html.svg?label=docs&down_message=unavailable&up_message=available" alt="Documentation Status">\n</a>\n<a href="https://opensource.org/licenses/MPL-2.0">\n<img src="https://img.shields.io/badge/License-MIT-yellow.svg" alt="License: MIT">\n</a>\n<a href="https://zenodo.org/badge/latestdoi/594445054"><img src="https://zenodo.org/badge/594445054.svg" alt="DOI"></a>\n\n`galilei` is a python package that makes emulating a numerical functions easier and more composable. It supports multiple backends such as pytorch-based neural networks, GPy-based gaussian process regression, etc. As of now, it defaults to a jax+flax+optax backend which supports automatic differenciation of the emulated function and easy composibility with the rest of the jax-based eco-system.\n\nThe motivation of emulating a function is that sometimes computing a function could be a time consuming task, so one may need to find fast approximations of a function that\'s better than basic interpolation techniques. Machine learning techniques such as neural networks offer a solution to this problem which is generic enough to emulate any arbitrarily shaped function. In contrast to the original function, a neural-network based emulator runs blazingly fast and even more so with GPU, often achieveing over many orders of magnitude speed-up.\n\nThis idea of emulating function is not new. In the field of cosmology we have powerful tools such as\n[cosmopower](https://github.com/alessiospuriomancini/cosmopower) and its derived works such as [axionEmu](https://github.com/keirkwame/axionEmu), whose idea inspired this work. My aim in this work differs from the previous approach, as I intend to make a both generic and easily-composible function emulator that can take any arbitrary parametrized numerical function as an input, and return a function with the exact same signature as a drop-in replacement in existing code base, with no additional code changes needed. In addition, I also focus on making the emulated function automatically differenciable regardless of its original implementation.\n\n## Features\n- Support multiple backends: `torch`, `sklearn`, `gpy` (for gaussian process regression), `jax`.\n- Flexible: Able to emulate generic numerical functions.\n- Automatic differenciable (supported by selected backends): emulated function is automatically differenciable and easily composible with jax-based tools.\n- Easy to use: just add a decorator `@emulate` and use your emulated function as a drop-in replacement of your existing function in code-base without additional modification.\n- Allow arbitrary transformation of function output before training through the use of `Preconditioner`.\n\n\n## Installation\n```\npip install galilei\n```\n\n## Basic usage\nSuppose that we have an expensive function that we want to emulate\n```python\ndef myfun(a, b):\n    # assume this is a slow function\n    x = np.linspace(0, 10, 100)\n    return np.sin(a*x) + np.sin(b*x)\n```\nIf you want to emulate this function, you can simply add a decorator `@emulate` and supply the parameters that you want to evaluate this function at to build up the training data set.\n\n```python\nfrom galilei import emulate\n\n@emulate(samples={\n    \'a\': np.random.rand(1000),\n    \'b\': np.random.rand(1000)\n})\ndef myfun(a, b):\n    x = np.linspace(0, 10, 100)\n    return np.sin(a*x) + np.sin(b*x)\n```\nHere we are just making 1000 pairs of random numbers from 0 to 1 to train our function. When executing these lines, the emulator will start training, and once it is done, the original `myfun` function will be automatically replaced with the emulated version and should behave in the same way, except much faster!\n```\nTraining emulator...\n100%|██████████| 500/500 [00:09<00:00, 50.50it/s, loss=0.023]\nAve Test loss: 0.025\n```\n![Comparison](https://github.com/guanyilun/galilei/raw/master/data/demo.png)\n\nWith the default `jax` backend, the emulated function is automatically jax compatible, which means one can easily compose them with `jax` machinary, such as in example below where I have compiled the emulated function with `jit` and then vectorized it over its first argument with `vmap`.\n```python\nfrom jax import jit, vmap\n\nvmap(jit(myfun), in_axes=(0, None))(np.linspace(0, 1, 10), 0.5)\n```\nOutput:\n```\nArray([[-2.39813775e-02,  2.16133818e-02,  8.05920288e-02,\n         1.66035295e-01,  2.01425016e-01,  2.42054626e-01,\n         2.74079561e-01,  3.50277930e-01,  4.12616253e-01,\n         4.33193207e-01,  4.82740909e-01,  5.66871405e-01,\n         5.73131263e-01,  6.51429832e-01,  6.55564785e-01,\n         ...\n```\nThe emulated function will also be automatically differenciable regardless of the original implementation details. For example, we could easily evaluate its jacobian (without finite differencing) with\n```python\nfrom jax import jacfwd\n\njacfwd(myfun, argnums=(0,1))(0.2, 0.8)\n```\nOutput:\n```\n(Array([ 0.05104188,  0.18436229,  0.08595917,  0.06582363,  0.17270228, ...],      dtype=float32),\n Array([-3.3511031e-01,  1.2647966e-01,  4.3209594e-02,  2.4325712e-01, ...],      dtype=float32))\n```\nYou can also easily save your trained model with the `save` option\n```python\n@emulate(samples={\n    \'a\': np.random.rand(100),\n    \'b\': np.random.rand(100)\n}, backend=\'sklearn\', save="test.pkl")\ndef myfun(a, b):\n    x = np.linspace(0, 10, 100)\n    return np.sin(a*x) + np.sin(b*x)\n```\nand when you use it in production, simply load a pretrained model with\n```python\n@emulate(backend=\'sklearn\', load="test.pkl")\ndef myfun(a, b):\n    x = np.linspace(0, 10, 100)\n    return np.sin(a*x) + np.sin(b*x)\n```\nand your function will be replaced with a fast emulated version.\n![Comparison2](https://github.com/guanyilun/galilei/raw/master/data/demo2.png)\n\nIt\'s also possible to sample training points based on latin hypercube using the `build_samples` function. For example, here I build a 100 sample latin hypercube for a given range of `a` and `b`\n```python\nfrom galilei.sampling import build_samples\n@emulate(\n    samples=build_samples({"a": [0, 2], "b": [0, 2]}, 100),\n    backend=\'sklearn\'\n)\ndef myfun(a, b):\n    x = np.linspace(0, 10, 100)\n    return np.sin(a*x) + np.sin(b*x)\n```\nSometimes one might want to collect training data only instead of training the emulator. This could\nbe done by\n```python\nfrom galilei.experimental import collect\n\n@collect(\n    samples=build_samples({"a": [0, 2], "b": [0, 2]}, 100),\n    save="collection.pkl",\n    mpi=True\n)\ndef myfun(a, b):\n    x = np.linspace(0, 10, 100)\n    return np.sin(a*x) + np.sin(b*x)\n```\nwhich will save a precomputed collection to `collection.pkl` for future loading. Note that the option to use `mpi` depends on the user having a working `mpi4py` which needs to be installed by the user.\nThe collection could be loaded for training emulator using\n```python\n@emulate(\n    collection="collection.pkl",\n    backend=\'sklearn\'\n)\ndef myfunc(a, b):\n    raise Exception()\n\nmyfunc(1, 1)\n```\nsince the function will not be evaluated in this case, we note that the implementation of `myfunc` makes no difference (otherwise it would have given an error).\n\nFor more usage examples, see this notebook:\n<a href="https://colab.research.google.com/drive/1_pvuAIqLUz4gV1vxytueb7AMR6Jmx-8n?usp=sharing">\n<img src="https://user-content.gitlab-static.net/dfbb2c197c959c47da3e225b71504edb540e21d6/68747470733a2f2f636f6c61622e72657365617263682e676f6f676c652e636f6d2f6173736574732f636f6c61622d62616467652e737667" alt="open in colab">\n</a>\n\n## Roadmap\n\n* TODO add prebuild preconditioners\n* TODO support downloading files from web\n* TODO auto infer backend\n* TODO chains of preconditioners\n\n## Credits\nThis package was created with the [ppw](https://zillionare.github.io/python-project-wizard) tool. For more information, please visit the [project page](https://zillionare.github.io/python-project-wizard/).\n\nIf this package is helpful in your work, please consider citing:\n```bibtex\n@article{yguan_2023,\n    title={galilei: a generic function emulator},\n    DOI={10.5281/zenodo.7651315},\n    publisher={Zenodo},\n    author={Yilun Guan},\n    year={2023},\n    month={Feb}}\n```\n\nFree software: MIT\n',
    'author': 'Yilun Guan',
    'author_email': 'zoom.aaron@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/guanyilun/galilei',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List

import jax
import jax.numpy as jnp
import optax
from flax import linen as nn
from tqdm import tqdm

from ..emulator import Emulator


class Net(nn.Module):
    """
    A class representing a simple fully-connected neural network.

    Parameters
    ----------
    odim: int
        Dimensionality of the output.
    hidden: list of int
        List of hidden layer sizes.
    """

    odim: int
    hidden: List[int]

    @nn.compact
    def __call__(self, x):
        for h in self.hidden:
            x = nn.Dense(h)(x)
            x = nn.sigmoid(x)
        x = nn.Dense(self.odim)(x)
        return x


class JaxEmulator(Emulator):
    def __init__(
        self,
        NNClass=Net,
        epochs=500,
        optimizer=None,
        lr=0.01,
        NN_kwargs={},
        batch_size=128,
        seed=42,
        **kwargs,
    ):
        """
        A class for emulating models using JAX.

        Parameters
        ----------
        NNClass: class, optional (default=Net)
            A class that implements the neural network architecture.
        epochs: int, optional (default=500)
            Number of training epochs.
        optimizer: str or callable, optional (default=None)
            Name of the optimizer or a custom optimizer function.
        lr: float, optional (default=0.01)
            Learning rate for the optimizer.
        NN_kwargs: dict, optional (default={})
            Additional keyword arguments to pass to the neural network constructor.
        batch_size: int, optional (default=128)
            Batch size for training.
        seed: int, optional (default=42)
            Seed for the random number generator.
        **kwargs: dict
            Additional keyword arguments to pass to the `Emulator` constructor.

        """
        super(JaxEmulator, self).__init__(**kwargs)
        self.NNClass = NNClass
        self.epochs = epochs
        self.optimizer = (
            optimizer if optimizer is not None else optax.adam(learning_rate=lr)
        )
        self.batch_size = batch_size
        self.seed = seed
        self.NN_kwargs = NN_kwargs

    def _train(self, X_train, Y_train):
        """
        Train the emulator

        Parameters
        ----------
        X_train : np.ndarray
            Training input
        Y_train : np.ndarray
            Training output
        """
        X_train = jnp.array(X_train)
        Y_train = jnp.array(Y_train)

        # update default NN kwargs if necessary
        NN_kwargs = {"odim": Y_train.shape[1], "hidden": [128, 128]}
        NN_kwargs.update(self.NN_kwargs)
        model = self.NNClass(**NN_kwargs)
        params = model.init(
            jax.random.PRNGKey(self.seed), jnp.ones((1, X_train.shape[1]))
        )
        opt_state = self.optimizer.init(params)

        # minibatch setup
        n_samples = X_train.shape[0]
        n_batches = n_samples // self.batch_size
        if n_batches > 0:
            X_train = jnp.split(X_train[: n_batches * self.batch_size], n_batches)
            Y_train = jnp.split(Y_train[: n_batches * self.batch_size], n_batches)
        else:
            X_train = [X_train]
            Y_train = [Y_train]

        # loss function and gradient for optimization
        @jax.jit
        def compute_loss(params, X, Y):
            loss = optax.l2_loss(model.apply(params, X), Y)
            loss = jnp.mean(loss)
            return loss

        val_and_grad = jax.jit(jax.value_and_grad(compute_loss))

        @jax.jit
        def train_step(params, opt_state, X_train, Y_train):
            for X, Y in zip(X_train, Y_train):
                loss_val, grads = val_and_grad(params, X, Y)
                updates, opt_state = self.optimizer.update(grads, opt_state)
                params = optax.apply_updates(params, updates)
            return params, opt_state, loss_val

        # training loop
        pbar = tqdm(range(self.epochs))
        for epoch in pbar:
            params, opt_state, loss_val = train_step(
                params, opt_state, X_train, Y_train
            )
            pbar.set_postfix({"loss": f"{loss_val:.5f}"})
        self.model = model
        self.params = params

    def _test(self, X_test, Y_test):
        """
        Test the emulator on a test set

        Parameters
        ----------
        X_test : array-like
            The test set input data
        Y_test : array-like
            The test set output data
        """
        X_test = jnp.array(X_test)
        Y_test = jnp.array(Y_test)
        Y_pred = self.model.apply(self.params, X_test)
        loss = jnp.mean(optax.l2_loss(Y_pred, Y_test))
        print(f"Test loss: {loss:.5f}")
        return loss

    def _predict(self, x):
        """
        Predict the output of the emulator for a given input

        Parameters
        ----------
        x : array-like
            The input data

        Returns
        -------
        The predicted output
        """
        return self.model.apply(self.params, x)

    def _save(self, store):
        """
        Save the emulator model to a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk

        """
        store["model"] = self.model
        store["params"] = self.params

    def _load(self, store):
        """
        Load the emulator model from a file

        Parameters
        ----------
        store: dict
            The dictionary that will be stored to disk
        """
        self.model = store["model"]
        self.params = store["params"]

    def _build_func(self):
        """
        Build a function that can be used to emulate the original function

        Returns
        -------
        emulated_func : callable
            Function that can be used to emulate the original function. The function has a
            reference to the emulator instance in the attribute `emulator`.
        """

        def emulated_func(*args, **kwargs):
            # check if the function has been called with the correct number of arguments
            if len(args) > 0:
                if len(args) != len(self.param_keys):
                    raise ValueError("Incorrect number of arguments")
                values = jnp.array(args)
            else:
                values = [kwargs.get(k, None) for k in self.param_keys]
                if None in values:
                    raise ValueError("Missing argument")
                values = jnp.array(values)
            res = self._predict(values)
            if self.preconditioner is not None:
                res = self.preconditioner.backward(res)
            return res

        # attach a reference to the emulator instance in the function
        emulated_func.emulator = self
        return emulated_func
